# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
2700 - Module for testing AQ with DbObject payloads.
"""

import decimal
import threading
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    book_type_name = "UDT_BOOK"
    book_queue_name = "TEST_BOOK_QUEUE"
    book_data = [
        ("Wings of Fire", "A.P.J. Abdul Kalam", decimal.Decimal("15.75")),
        ("The Story of My Life", "Hellen Keller", decimal.Decimal("10.50")),
        ("The Chronicles of Narnia", "C.S. Lewis", decimal.Decimal("25.25")),
    ]

    def __deq_in_thread(self, results):
        with test_env.get_connection() as conn:
            books_type = conn.gettype(self.book_type_name)
            queue = conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.wait = 10
            props = queue.deqone()
            if props is not None:
                book = props.payload
                results.append((book.TITLE, book.AUTHORS, book.PRICE))
            conn.commit()

    def __verify_attr(self, obj, attrName, value):
        setattr(obj, attrName, value)
        self.assertEqual(getattr(obj, attrName), value)

    def test_2700(self):
        "2700 - test dequeuing an empty queue"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)

    def test_2701(self):
        "2701 - test enqueuing and dequeuing multiple messages"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        props = self.conn.msgproperties()
        for title, authors, price in self.book_data:
            props.payload = book = queue.payload_type.newobject()
            book.TITLE = title
            book.AUTHORS = authors
            book.PRICE = price
            queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            book = props.payload
            row = (book.TITLE, book.AUTHORS, book.PRICE)
            results.append(row)
        self.conn.commit()
        self.assertEqual(results, self.book_data)

    def test_2702(self):
        "2702 - test dequeuing with DEQ_REMOVE_NODATA option"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[1]
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA
        props = queue.deqone()
        self.assertIsNotNone(props)
        self.assertIsNone(props.payload.TITLE)

    def test_2703(self):
        "2703 - test getting/setting dequeue options attributes"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        options = queue.deqoptions
        self.__verify_attr(options, "condition", "TEST_CONDITION")
        self.__verify_attr(options, "consumername", "TEST_CONSUMERNAME")
        self.__verify_attr(options, "correlation", "TEST_CORRELATION")
        self.__verify_attr(options, "mode", oracledb.DEQ_LOCKED)
        self.__verify_attr(
            options, "navigation", oracledb.DEQ_NEXT_TRANSACTION
        )
        self.__verify_attr(options, "transformation", "TEST_TRANSFORMATION")
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)
        self.__verify_attr(options, "wait", 1287)
        self.__verify_attr(options, "msgid", b"mID")

    def test_2704(self):
        "2704 - test waiting for dequeue"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        book = queue.payload_type.newobject()
        title, authors, price = self.book_data[0]
        book.TITLE = title
        book.AUTHORS = authors
        book.PRICE = price
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        self.conn.commit()
        thread.join()
        self.assertEqual(results, [(title, authors, price)])

    def test_2705(self):
        "2705 - test getting/setting enqueue options attributes"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        options = queue.enqoptions
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)

    def test_2706(self):
        "2706 - test errors for invalid values for enqueue"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        self.assertRaises(TypeError, queue.enqone, book)

    def test_2707(self):
        "2707 - test getting/setting message properties attributes"
        props = self.conn.msgproperties()
        self.__verify_attr(props, "correlation", "TEST_CORRELATION")
        self.__verify_attr(props, "delay", 60)
        self.__verify_attr(props, "exceptionq", "TEST_EXCEPTIONQ")
        self.__verify_attr(props, "expiration", 30)
        self.assertEqual(props.attempts, 0)
        self.__verify_attr(props, "priority", 1)
        self.assertEqual(props.state, oracledb.MSG_READY)
        self.assertEqual(props.deliverymode, 0)

    def test_2708(self):
        "2708 - test enqueue visibility option - ENQ_ON_COMMIT"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)
        self.conn.commit()
        props = queue.deqone()
        other_conn.commit()
        self.assertIsNotNone(props)

    def test_2709(self):
        "2709 - test enqueue visibility option - ENQ_IMMEDIATE"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_conn.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2710(self):
        "2710 - test enqueue/dequeue delivery modes identical - buffered"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_conn.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2711(self):
        "2711 - test enqueue/dequeue delivery modes identical - persistent"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_conn.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2712(self):
        "2712 - test enqueue/dequeue delivery modes the same"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_conn.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2713(self):
        "2713 - test enqueue/dequeue delivery modes different"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)

    @unittest.skipIf(
        test_env.get_is_thin(), "Thin mode doesn't support transformation yet"
    )
    def test_2714(self):
        "2714 - test dequeue transformation"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        expected_price = book.PRICE + 10
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        self.conn.commit()

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        transformation_str = f"{self.conn.username}.transform2"
        queue.deqoptions.transformation = transformation_str
        self.assertEqual(queue.deqoptions.transformation, transformation_str)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertEqual(props.payload.PRICE, expected_price)

    @unittest.skipIf(
        test_env.get_is_thin(), "Thin mode doesn't support transformation yet"
    )
    def test_2715(self):
        "2715 - test enqueue transformation"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        expected_price = book.PRICE + 5
        queue.enqoptions.transformation = transformation_str = (
            f"{self.conn.username}.transform1"
        )
        queue.enqoptions.transformation = transformation_str
        if test_env.get_client_version() >= (23, 1):
            self.assertEqual(
                queue.enqoptions.transformation, transformation_str
            )
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        self.conn.commit()

        other_conn = test_env.get_connection()
        books_type = other_conn.gettype(self.book_type_name)
        queue = other_conn.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertEqual(props.payload.PRICE, expected_price)

    def test_2716(self):
        "2716 - test to verify payloadType is deprecated"
        books_type = self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, payloadType=books_type)
        self.assertEqual(queue.payload_type, books_type)
        self.assertEqual(queue.payloadType, books_type)
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.queue(
                self.book_queue_name, books_type, payloadType=books_type
            )

    def test_2717(self):
        "2717 - test error for message with no payload"
        books_type = self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        props = self.conn.msgproperties()
        with self.assertRaisesFullCode("DPY-2000"):
            queue.enqone(props)

    def test_2718(self):
        "2718 - verify that the msgid property is returned correctly"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        props = self.conn.msgproperties(payload=book)
        self.assertIsNone(props.msgid)
        queue.enqone(props)
        self.cursor.execute("select msgid from book_queue_tab")
        (actual_msgid,) = self.cursor.fetchone()
        self.assertEqual(props.msgid, actual_msgid)
        props = queue.deqone()
        self.assertEqual(props.msgid, actual_msgid)

    @unittest.skipIf(
        test_env.get_is_thin(), "Thin mode doesn't support recipient list yet"
    )
    def test_2719(self):
        "2719 - verify use of recipients property"
        books_type = self.conn.gettype(self.book_type_name)
        book = books_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue = self.conn.queue("BOOK_QUEUE_MULTI", books_type)
        props = self.conn.msgproperties(
            payload=book, recipients=["sub2", "sub3"]
        )
        self.assertEqual(props.recipients, ["sub2", "sub3"])
        queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.consumername = "sub3"
        props1 = queue.deqone()
        book = props1.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        self.assertEqual(results, self.book_data[0])
        queue.deqoptions.consumername = "sub1"
        props1 = queue.deqone()
        self.assertIsNone(props1)

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support notification yet"
    )
    def test_2720(self):
        "2720 - verify attributes of AQ message which spawned notification"
        if self.is_on_oracle_cloud(self.conn):
            self.skipTest("AQ notification not supported on the cloud")
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        condition = threading.Condition()
        conn = test_env.get_connection(events=True)

        def notification_callback(message):
            self.cursor.execute("select msgid from book_queue_tab")
            (actual_msgid,) = self.cursor.fetchone()
            self.assertEqual(message.msgid, actual_msgid)
            self.assertIsNone(message.consumer_name)
            main_user = test_env.get_main_user().upper()
            self.assertEqual(
                message.queue_name, f'"{main_user}"."{queue.name}"'
            )
            self.assertEqual(message.type, oracledb.EVENT_AQ)
            with condition:
                condition.notify()

        sub = conn.subscribe(
            namespace=oracledb.SUBSCR_NAMESPACE_AQ,
            name=self.book_queue_name,
            timeout=300,
            callback=notification_callback,
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        self.conn.commit()
        with condition:
            self.assertTrue(condition.wait(5))
        conn.unsubscribe(sub)

    def test_2721(self):
        "2721 - test message props enqtime"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        self.cursor.execute("select current_timestamp from dual")
        (start_date,) = self.cursor.fetchone()
        start_date = start_date.replace(microsecond=0)
        props = self.conn.msgproperties(payload=book)
        queue.enqone(props)
        props = queue.deqone()
        self.cursor.execute("select current_timestamp from dual")
        (end_date,) = self.cursor.fetchone()
        end_date = end_date.replace(microsecond=0)
        self.assertTrue(start_date <= props.enqtime <= end_date)

    def test_2722(self):
        "2722 - test message props declared attributes"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        values = dict(
            payload=book,
            correlation="TEST_CORRELATION",
            delay=7,
            exceptionq="TEST_EXCEPTIONQ",
            expiration=10,
            priority=1,
        )
        props = self.conn.msgproperties(**values)
        for attr_name in values:
            self.assertEqual(getattr(props, attr_name), values[attr_name])

    def test_2723(self):
        "2723 - test error for invalid type for payload_type"
        self.assertRaises(
            TypeError, self.conn.queue, "THE QUEUE", payload_type=4
        )

    def test_2724(self):
        "2724 - test setting bytes to payload"
        props = self.conn.msgproperties()
        bytes_val = b"Hello there"
        props.payload = bytes_val
        self.assertEqual(props.payload, bytes_val)

    def test_2725(self):
        "2725 - test getting queue attributes"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        self.assertEqual(queue.name, self.book_queue_name)
        self.assertEqual(queue.connection, self.conn)

    def test_2726(self):
        "2726 - test getting write-only attributes"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        with self.assertRaises(AttributeError):
            queue.enqoptions.deliverymode
        with self.assertRaises(AttributeError):
            queue.deqoptions.deliverymode

    def test_2727(self):
        "2727 - test correlation deqoption"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        correlations = ["Math", "Programming"]
        num_messages = 3
        messages = [
            self.conn.msgproperties(payload=book, correlation=c)
            for c in correlations
            for i in range(num_messages)
        ]
        queue.enqmany(messages)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.correlation = correlations[0]
        correlated_messages = queue.deqmany(num_messages + 1)
        self.assertEqual(len(correlated_messages), num_messages)

        queue.deqoptions.correlation = correlations[1]
        with self.assertRaisesFullCode("ORA-25241"):
            queue.deqone()
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        correlated_messages = queue.deqmany(num_messages + 1)
        self.assertEqual(len(correlated_messages), num_messages)

    def test_2728(self):
        "2728 - test correlation deqoption with pattern-matching characters"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        for correlation in ("PreCalculus-math1", "Calculus-Math2"):
            props = self.conn.msgproperties(
                payload=book, correlation=correlation
            )
            queue.enqone(props)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.correlation = "%Calculus-%ath_"
        messages = queue.deqmany(5)
        self.assertEqual(len(messages), 2)

    def test_2729(self):
        "2729 - test condition deqoption with priority"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT

        priorities = [5, 10]
        indexes = [0, 1]
        for priority, ix in zip(priorities, indexes):
            book = queue.payload_type.newobject()
            book.TITLE, book.AUTHORS, book.PRICE = self.book_data[ix]
            props = self.conn.msgproperties(payload=book, priority=priority)
            queue.enqone(props)

        queue.deqoptions.condition = "priority = 9"
        messages = queue.deqmany(3)
        self.assertEqual(len(messages), 0)

        for priority, ix in zip(priorities, indexes):
            queue.deqoptions.condition = f"priority = {priority}"
            messages = queue.deqmany(3)
            self.assertEqual(len(messages), 1)
            book = messages[0].payload
            data = book.TITLE, book.AUTHORS, book.PRICE
            self.assertEqual(data, self.book_data[ix])

    def test_2730(self):
        "2730 - test mode deqoption with DEQ_REMOVE_NODATA"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA

        book = queue.payload_type.newobject()
        for data in self.book_data:
            book.TITLE, book.AUTHORS, book.PRICE = data
            props = self.conn.msgproperties(payload=book)
            queue.enqone(props)

        messages = queue.deqmany(5)
        self.assertEqual(len(messages), 3)
        for message in messages:
            self.assertIsNone(message.payload.TITLE)
            self.assertIsNone(message.payload.AUTHORS)
            self.assertIsNone(message.payload.PRICE)

    def test_2731(self):
        "2731 - test payload_type returns the correct value"
        books_type = self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        self.assertEqual(queue.payload_type, books_type)

    def test_2732(self):
        "2732 - test deprecated attributes (enqOptions, deqOptions)"
        books_type = self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        self.assertEqual(queue.enqOptions, queue.enqoptions)
        self.assertEqual(queue.deqOptions, queue.deqoptions)

    def test_2733(self):
        "2733 - test deprecated AQ methods (enqOne, deqOne)"
        books_type = self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqOne(self.conn.msgproperties(book))
        props = queue.deqOne()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        self.assertEqual(results, self.book_data[0])

    def test_2734(self):
        "2734 - test enqueuing to an object queue with the wrong payload"
        queue = self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        props = self.conn.msgproperties(payload="A string")
        with self.assertRaisesFullCode("DPY-2062"):
            queue.enqone(props)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5800 - Module for testing cursor variables with asyncio
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_5800(self):
        "5800 - test binding in a cursor"
        cursor = self.conn.cursor()
        self.assertIsNone(cursor.description)
        await self.cursor.execute(
            """
            begin
                open :cursor for select 'X' StringValue from dual;
            end;
            """,
            cursor=cursor,
        )
        varchar_ratio, _ = await test_env.get_charset_ratios_async()
        expected_value = [
            (
                "STRINGVALUE",
                oracledb.DB_TYPE_CHAR,
                1,
                varchar_ratio,
                None,
                None,
                True,
            )
        ]
        self.assertEqual(cursor.description, expected_value)
        self.assertEqual(await cursor.fetchall(), [("X",)])

    async def test_5801(self):
        "5801 - test binding in a cursor from a package"
        cursor = self.conn.cursor()
        self.assertIsNone(cursor.description)
        await self.cursor.callproc(
            "pkg_TestRefCursors.TestOutCursor", (2, cursor)
        )
        varchar_ratio, _ = await test_env.get_charset_ratios_async()
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            (
                "STRINGCOL",
                oracledb.DB_TYPE_VARCHAR,
                20,
                20 * varchar_ratio,
                None,
                None,
                False,
            ),
        ]
        self.assertEqual(cursor.description, expected_value)
        self.assertEqual(
            await cursor.fetchall(), [(1, "String 1"), (2, "String 2")]
        )

    async def test_5802(self):
        "5802 - test that binding the cursor itself is not supported"
        cursor = self.conn.cursor()
        sql = """
                begin
                    open :pcursor for
                        select 1 from dual;
                end;"""
        with self.assertRaisesFullCode("DPY-3009"):
            await cursor.execute(sql, pcursor=cursor)

    async def test_5803(self):
        "5803 - test returning a ref cursor after closing it"
        out_cursor = self.conn.cursor()
        sql = """
                begin
                    open :pcursor for
                        select IntCol
                        from TestNumbers
                        order by IntCol;
                end;"""
        await self.cursor.execute(sql, pcursor=out_cursor)
        rows = await out_cursor.fetchall()
        out_cursor.close()
        out_cursor = self.conn.cursor()
        await self.cursor.execute(sql, pcursor=out_cursor)
        rows2 = await out_cursor.fetchall()
        self.assertEqual(rows, rows2)

    async def test_5804(self):
        "5804 - test fetching a cursor"
        await self.cursor.execute(
            """
            select IntCol, cursor(select IntCol + 1 from dual) CursorValue
            from TestNumbers
            order by IntCol
            """
        )
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            (
                "CURSORVALUE",
                oracledb.DB_TYPE_CURSOR,
                None,
                None,
                None,
                None,
                True,
            ),
        ]
        self.assertEqual(self.cursor.description, expected_value)
        for i in range(1, 11):
            number, cursor = await self.cursor.fetchone()
            self.assertEqual(number, i)
            self.assertEqual(await cursor.fetchall(), [(i + 1,)])

    async def test_5805(self):
        "5805 - test that ref cursor binds cannot use optimised path"
        ref_cursor = self.conn.cursor()
        sql = """
                begin
                    open :rcursor for
                        select IntCol, StringCol
                        from TestStrings where IntCol
                        between :start_value and :end_value;
                end;"""
        expected_value = [(2, "String 2"), (3, "String 3"), (4, "String 4")]
        await self.cursor.execute(
            sql, rcursor=ref_cursor, start_value=2, end_value=4
        )
        self.assertEqual(await ref_cursor.fetchall(), expected_value)
        ref_cursor.close()

        expected_value = [(5, "String 5"), (6, "String 6")]
        ref_cursor = self.conn.cursor()
        await self.cursor.execute(
            sql, rcursor=ref_cursor, start_value=5, end_value=6
        )
        self.assertEqual(await ref_cursor.fetchall(), expected_value)

    async def test_5806(self):
        "5806 - test round trips using a REF cursor"
        await self.setup_round_trip_checker()

        # simple DDL only requires a single round trip
        with self.conn.cursor() as cursor:
            await cursor.execute("truncate table TestTempTable")
            await self.assertRoundTrips(1)

        # array execution only requires a single round trip
        num_rows = 590
        with self.conn.cursor() as cursor:
            sql = "insert into TestTempTable (IntCol) values (:1)"
            data = [(n + 1,) for n in range(num_rows)]
            await cursor.executemany(sql, data)
            await self.assertRoundTrips(1)

        # create REF cursor and execute stored procedure
        # (array size set before procedure is called)
        with self.conn.cursor() as cursor:
            refcursor = self.conn.cursor()
            refcursor.arraysize = 150
            await cursor.callproc("myrefcursorproc", [refcursor])
            await refcursor.fetchall()
            await self.assertRoundTrips(5)

        # create REF cursor and execute stored procedure
        # (array size set after procedure is called)
        with self.conn.cursor() as cursor:
            refcursor = self.conn.cursor()
            await cursor.callproc("myrefcursorproc", [refcursor])
            refcursor.arraysize = 145
            await refcursor.fetchall()
            await self.assertRoundTrips(6)

    async def test_5807(self):
        "5807 - test executing different SQL after getting a REF cursor"
        with self.conn.cursor() as cursor:
            refcursor = self.conn.cursor()
            await cursor.callproc("myrefcursorproc", [refcursor])
            var = cursor.var(int)
            await refcursor.execute("begin :1 := 15; end;", [var])
            self.assertEqual(var.getvalue(), 15)

    async def test_5808(self):
        "5808 - test calling a function that returns a REF cursor"
        with self.conn.cursor() as cursor:
            ref_cursor = await cursor.callfunc(
                "pkg_TestRefCursors.TestReturnCursor",
                oracledb.DB_TYPE_CURSOR,
                [2],
            )
            self.assertEqual(
                await ref_cursor.fetchall(), [(1, "String 1"), (2, "String 2")]
            )

    async def test_5809(self):
        "5809 - test using an output type handler with a REF cursor"

        def type_handler(cursor, metadata):
            return cursor.var(str, arraysize=cursor.arraysize)

        self.conn.outputtypehandler = type_handler
        var = self.cursor.var(oracledb.DB_TYPE_CURSOR)
        string_val = "Test String - 5809"
        with self.conn.cursor() as cursor:
            await cursor.callproc(
                "pkg_TestRefCursors.TestLobCursor", [string_val, var]
            )
        ref_cursor = var.getvalue()
        self.assertEqual(await ref_cursor.fetchall(), [(string_val,)])

    async def test_5810(self):
        "5810 - bind a REF cursor but never open it"
        ref_cursor_var = self.cursor.var(oracledb.DB_TYPE_CURSOR)
        await self.cursor.execute(
            """
            begin
                if false then
                    open :cursor for
                        select user
                        from dual;
                end if;
            end;
            """,
            cursor=ref_cursor_var,
        )
        ref_cursor = ref_cursor_var.getvalue()
        if ref_cursor is not None:
            with self.assertRaisesFullCode("DPY-4025"):
                await ref_cursor.fetchall()

    async def test_5811(self):
        "5811 - test fetching a cursor with a custom class"

        class Counter:
            num_cursors_created = 0

            @classmethod
            def cursor_created(cls):
                cls.num_cursors_created += 1

        class MyConnection(oracledb.AsyncConnection):
            def cursor(self):
                Counter.cursor_created()
                return super().cursor()

        conn = await test_env.get_connection_async(conn_class=MyConnection)
        cursor = conn.cursor()
        await cursor.execute(
            """
            select
                cursor(select 1 from dual),
                cursor(select 2 from dual)
            from dual
            """
        )
        await cursor.fetchall()
        self.assertEqual(Counter.num_cursors_created, 3)

    async def test_5812(self):
        "5812 - test that nested cursors are fetched correctly"
        sql = """
            select
            'Level 1 String',
            cursor(
                select
                'Level 2 String',
                cursor(
                    select
                    'Level 3 String',
                    cursor(
                        select 1, 'Level 4 String A' from dual
                        union all
                        select 2, 'Level 4 String B' from dual
                        union all
                        select 3, 'Level 4 String C' from dual
                    ) as nc3
                    from dual
                ) as nc2
                from dual
            ) as nc1
            from dual"""
        await self.cursor.execute(sql)

        async def transform_row(r):
            return tuple([await transform_fn(v) for v in r])

        async def transform_fn(v):
            if isinstance(v, oracledb.AsyncCursor):
                return [await transform_row(r) async for r in v]
            return v

        rows = [await transform_row(r) async for r in self.cursor]
        expected_value = [
            (
                "Level 1 String",
                [
                    (
                        "Level 2 String",
                        [
                            (
                                "Level 3 String",
                                [
                                    (1, "Level 4 String A"),
                                    (2, "Level 4 String B"),
                                    (3, "Level 4 String C"),
                                ],
                            ),
                        ],
                    ),
                ],
            )
        ]
        self.assertEqual(rows, expected_value)

    async def test_5813(self):
        "5813 - test fetching nested cursors with more columns than parent"
        sql = """
            select
                'Top Level String',
                cursor(
                    select
                        'Nested String 1',
                        'Nested String 2',
                        'Nested String 3'
                    from dual
                )
            from dual"""
        await self.cursor.execute(sql)

        async def transform_row(r):
            return tuple([await transform_fn(v) for v in r])

        async def transform_fn(v):
            if isinstance(v, oracledb.AsyncCursor):
                return [await transform_row(r) async for r in v]
            return v

        rows = [await transform_row(r) async for r in self.cursor]
        expected_value = [
            (
                "Top Level String",
                [("Nested String 1", "Nested String 2", "Nested String 3")],
            )
        ]
        self.assertEqual(rows, expected_value)

    async def test_5814(self):
        "5814 - test reusing a closed ref cursor for executing different sql"
        sql = "select 58141, 'String 58141' from dual"
        ref_cursor = self.conn.cursor()
        ref_cursor.prefetchrows = 0
        await ref_cursor.execute(sql)
        plsql = "begin pkg_TestRefCursors.TestCloseCursor(:rcursor); end;"
        await self.cursor.execute(plsql, rcursor=ref_cursor)
        sql = "select 58142, 'String 58142' from dual"
        await ref_cursor.execute(sql)
        self.assertEqual(
            await ref_cursor.fetchall(),
            [
                (58142, "String 58142"),
            ],
        )

    async def test_5815(self):
        "5815 - test reusing a closed ref cursor for executing same sql"
        sql = "select 5815, 'String 5815' from dual"
        ref_cursor = self.conn.cursor()
        ref_cursor.prefetchrows = 0
        await ref_cursor.execute(sql)
        plsql = "begin pkg_TestRefCursors.TestCloseCursor(:rcursor); end;"
        await self.cursor.execute(plsql, rcursor=ref_cursor)
        await ref_cursor.execute(sql)
        self.assertEqual(
            await ref_cursor.fetchall(),
            [
                (5815, "String 5815"),
            ],
        )


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Module for testing dataframes
"""
import datetime
import decimal
import unittest

import oracledb

try:
    import pyarrow
    import pandas

    HAS_INTEROP = True
except ImportError:
    HAS_INTEROP = False

import test_env

# basic
DATASET_1 = [
    (
        1,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1989, 8, 22),
        12132.40,
        400,
        datetime.datetime.now(),
    ),
    (
        2,
        "Big",
        "Hero",
        "San Fransokyo",
        "Japansa",
        datetime.date(1988, 8, 22),
        234234.32,
        400,
        datetime.datetime.now(),
    ),
]

# None, -ve
DATASET_2 = [
    (
        1,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1989, 8, 22),
        None,
        400,
        datetime.datetime.now(),
    ),
    (
        2,
        "Big",
        "Hero",
        "San Fransokyo",
        None,
        datetime.date(1988, 8, 22),
        -12312.1,
        0,
        datetime.datetime.now(),
    ),
]

# None, +/- 0.XXX
DATASET_3 = [
    (
        1,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1989, 8, 22),
        None,
        400,
        datetime.datetime.now(),
    ),
    (
        2,
        "Big",
        "Hero",
        "San Fransokyo",
        None,
        datetime.date(1988, 8, 22),
        0.12,
        0,
        datetime.datetime.now(),
    ),
    (
        3,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1989, 8, 22),
        None,
        400,
        datetime.datetime.now(),
    ),
    (
        4,
        "Big",
        "Hero",
        "San Fransokyo",
        None,
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
]

# Duplicates
DATASET_4 = [
    (
        1,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1989, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
    (
        2,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
    (
        3,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
    (
        4,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
    (
        5,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
    (
        6,
        "John",
        "Doe",
        "San Francisco",
        "USA",
        datetime.date(1988, 8, 22),
        -0.01,
        0,
        datetime.datetime.now(),
    ),
]


class TestCase(test_env.BaseTestCase):

    def __check_interop(self):
        """
        Checks to see if the pyarrow and pandas modules are available.
        """
        if not HAS_INTEROP:
            self.skipTest("missing pandas or pyarrow modules")

    def __convert_to_array(self, data, typ):
        """
        Convert raw data to an Arrow array using pyarrow.
        """
        if isinstance(typ, pyarrow.Decimal128Type):
            data = [
                decimal.Decimal(str(value)) if value is not None else value
                for value in data
            ]
        elif isinstance(typ, pyarrow.TimestampType):
            if typ.unit == "s":
                data = [
                    datetime.datetime(v.year, v.month, v.day).timestamp()
                    for v in data
                ]
            else:
                data = [value.timestamp() * 1000000 for value in data]
        mask = [value is None for value in data]
        return pyarrow.array(data, typ, mask=mask)

    def __convert_to_df(self, data):
        """
        Converts the data set to a Pandas data frame for comparison to what is
        returned from the database.
        """
        data_by_col = [[row[i] for row in data] for i in range(len(data[0]))]
        fetch_decimals = oracledb.defaults.fetch_decimals
        types = [
            pyarrow.decimal128(9) if fetch_decimals else pyarrow.int64(),
            pyarrow.string(),
            pyarrow.string(),
            pyarrow.string(),
            pyarrow.string(),
            pyarrow.timestamp("s"),
            pyarrow.decimal128(9, 2) if fetch_decimals else pyarrow.float64(),
            pyarrow.decimal128(3) if fetch_decimals else pyarrow.int64(),
            pyarrow.timestamp("us"),
        ]
        arrays = [
            self.__convert_to_array(d, t) for d, t in zip(data_by_col, types)
        ]
        names = [
            "ID",
            "FIRSTNAME",
            "LASTNAME",
            "CITY",
            "COUNTRY",
            "DATEOFBIRTH",
            "SALARY",
            "CREDITSCORE",
            "LASTUPDATED",
        ]
        pa_tab = pyarrow.Table.from_arrays(arrays, names=names)
        pa_tab.validate(full=True)
        return pa_tab.to_pandas()

    def __get_data_from_df(self, df):
        """
        Returns data from the data frame in a normalized fashion suitable for
        comparison. In particular, NaN values cannot be compared to one another
        so they are converted to the value None for comparison purposes.
        """
        return [
            tuple(None if pandas.isna(v) else v for v in row)
            for row in df.itertuples(index=False, name=None)
        ]

    def __populate_table(self, data):
        """
        Populate the test table with the given data.
        """
        self.cursor.execute("truncate table TestDataframe")
        types = [None] * len(data[0])
        types[8] = oracledb.DB_TYPE_TIMESTAMP
        self.cursor.setinputsizes(*types)
        self.cursor.executemany(
            """
            insert into TestDataframe (
                Id, FirstName, LastName, City, Country,
                DateOfBirth, Salary, CreditScore, LastUpdated
            ) values (
                :id, :first_name, :last_name, :city, :country,
                :dob, :salary, :credit_score, :last_updated
            )
            """,
            data,
        )
        self.conn.commit()

    def __test_df_interop(self, data):
        """
        Tests interoperability with external data frames using the data set
        provided.
        """
        self.__check_interop()
        self.__populate_table(data)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        self.__validate_df(ora_df, data)

    def __test_df_batches_interop(self, data, batch_size, num_batches):
        """
        Tests interoperability with external data frames using the data set
        provided.
        """
        self.__check_interop()
        self.__populate_table(data)
        statement = "select * from TestDataFrame order by Id"
        batches = list(self.conn.fetch_df_batches(statement, size=batch_size))
        self.assertEqual(len(batches), num_batches)
        if num_batches == 1:
            self.__validate_df(batches[0], data)
        else:
            offset = 0
            for batch in batches:
                self.__validate_df(batch, data[offset : offset + batch_size])
                offset += batch_size

    def __validate_df(self, ora_df, data):
        """
        Validates the data frame by converting it to Pandas and comparing it
        with the original data set that was used.
        """
        raw_df = self.__convert_to_df(data)
        raw_data = self.__get_data_from_df(raw_df)
        fetched_tab = pyarrow.Table.from_arrays(
            ora_df.column_arrays(), names=ora_df.column_names()
        )
        fetched_df = fetched_tab.to_pandas()
        fetched_data = self.__get_data_from_df(fetched_df)
        self.assertEqual(fetched_data, raw_data)

    def test_8000(self):
        "8000 - test basic fetch of data frame"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        self.assertEqual(ora_df.num_rows(), len(DATASET_1))
        self.assertEqual(ora_df.num_columns(), len(DATASET_1[0]))
        metadata = dict(
            num_columns=ora_df.num_columns(),
            num_rows=ora_df.num_rows(),
            num_chunks=1,
        )
        self.assertEqual(ora_df.metadata, metadata)

    def test_8001(self):
        "8001 - test conversion to external dataframe"
        self.__test_df_interop(DATASET_1)

    def test_8002(self):
        "8001 - test null and negative values"
        self.__test_df_interop(DATASET_2)

    def test_8003(self):
        "8002 - test with fetch_decimals"
        with test_env.DefaultsContextManager("fetch_decimals", True):
            self.__test_df_interop(DATASET_1)

    def test_8004(self):
        "8003 - test null and negative values with fetch_decimals"
        with test_env.DefaultsContextManager("fetch_decimals", True):
            self.__test_df_interop(DATASET_2)

    def test_8005(self):
        "8005 - test null and values with leading zeros"
        self.__test_df_interop(DATASET_3)

    def test_8006(self):
        "8005 - test null and values with leading zeros with fetch_decimals"
        with test_env.DefaultsContextManager("fetch_decimals", True):
            self.__test_df_interop(DATASET_3)

    def test_8007(self):
        "8007 - duplicate values in the rows"
        self.__test_df_interop(DATASET_4)

    def test_8008(self):
        "8008 - batches without specification of size"
        self.__test_df_batches_interop(
            DATASET_4, batch_size=None, num_batches=1
        )

    def test_8009(self):
        "8009 - batches with specification of size"
        self.__test_df_batches_interop(DATASET_4, batch_size=5, num_batches=2)

    def test_8010(self):
        "8010 - verify passing Arrow arrays twice works"
        self.__check_interop()
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        self.__validate_df(ora_df, DATASET_1)
        self.__validate_df(ora_df, DATASET_1)

    def test_8011(self):
        "8011 - verify empty data set"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame where Id = 4"
        ora_df = self.conn.fetch_df_all(statement)
        self.assertEqual(ora_df.num_rows(), 0)

    def test_8012(self):
        "8012 - verify empty data set with batches"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame where Id = 4"
        for ora_df in self.conn.fetch_df_batches(statement):
            self.assertEqual(ora_df.num_rows(), 0)

    def test_8013(self):
        "8013 - negative checks on attributes"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        with self.assertRaises(IndexError):
            ora_df.get_column(121)
        with self.assertRaises(IndexError):
            ora_df.get_column(-1)
        with self.assertRaises(KeyError):
            ora_df.get_column_by_name("missing_column")

    def test_8014(self):
        "8014 - check size and null count with no nulls"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        col = ora_df.get_column(0)
        self.assertEqual(col.size(), len(DATASET_1))
        self.assertEqual(col.null_count, 0)

    def test_8015(self):
        "8015 - check size and null count with nulls present"
        self.__populate_table(DATASET_2)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        col = ora_df.get_column_by_name("SALARY")
        self.assertEqual(col.size(), len(DATASET_2))
        self.assertEqual(col.null_count, 1)

    def test_8016(self):
        "8016 - check unsupported error"
        statement = "select cursor(select user from dual) from dual"
        with self.assertRaisesFullCode("DPY-3030"):
            self.conn.fetch_df_all(statement)

    def test_8017(self):
        "8017 - batches with specification of size matching number of rows"
        self.__test_df_batches_interop(
            DATASET_2, batch_size=len(DATASET_2), num_batches=1
        )

    def test_8018(self):
        "8018 - verify get_column() returns the correct value"
        self.__check_interop()
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        array = pyarrow.array(ora_df.get_column(1))
        self.assertEqual(array.to_pylist(), ["John", "Big"])

    def test_8019(self):
        "8019 - verify OracleColumn and get_buffers"
        self.__populate_table(DATASET_1)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)
        ora_col = ora_df.get_column(1)
        self.assertEqual(ora_col.num_chunks(), 1)
        self.assertEqual(ora_col.size(), 2)

        buffers = ora_col.get_buffers()
        self.assertEqual(len(buffers), 3)
        self.assertIsNotNone(buffers["data"])
        self.assertIsNotNone(buffers["offsets"])
        self.assertIsNone(buffers["validity"])

    def test_8020(self):
        "8020 - verify  OracleColumn Attributes"
        self.__populate_table(DATASET_2)
        statement = "select * from TestDataFrame order by Id"
        ora_df = self.conn.fetch_df_all(statement)

        ora_col = ora_df.get_column(0)
        self.assertEqual(ora_col.describe_null[0], 0)
        self.assertEqual(ora_col.dtype[0], 0)
        metadata = {"name": "ID", "size": 2, "num_chunks": 1}
        self.assertEqual(metadata, ora_col.metadata)
        self.assertEqual(ora_col.null_count, 0)

        ora_col = ora_df.get_column(4)
        self.assertEqual(ora_col.describe_null[0], 3)
        self.assertEqual(ora_col.dtype[0], 21)
        metadata = {"name": "COUNTRY", "size": 2, "num_chunks": 1}
        self.assertEqual(metadata, ora_col.metadata)
        self.assertEqual(ora_col.null_count, 1)

    def test_8021(self):
        "8021 - batches with size that has duplicate rows across batches"
        self.__test_df_batches_interop(DATASET_4, batch_size=3, num_batches=2)

    def test_8022(self):
        "8022 - fetch_decimals without precision and scale specified"
        data = [(1.0,)]
        self.__check_interop()
        with test_env.DefaultsContextManager("fetch_decimals", True):
            ora_df = self.conn.fetch_df_all("select 1.0 from dual")
            fetched_tab = pyarrow.Table.from_arrays(
                ora_df.column_arrays(), names=ora_df.column_names()
            )
            fetched_df = fetched_tab.to_pandas()
            fetched_data = self.__get_data_from_df(fetched_df)
            self.assertEqual(fetched_data, data)

    def test_8023(self):
        "8023 - fetch clob"
        data = [("test_8023",)]
        self.__check_interop()
        ora_df = self.conn.fetch_df_all(
            "select to_clob('test_8023') from dual"
        )
        fetched_tab = pyarrow.Table.from_arrays(
            ora_df.column_arrays(), names=ora_df.column_names()
        )
        fetched_df = fetched_tab.to_pandas()
        fetched_data = self.__get_data_from_df(fetched_df)
        self.assertEqual(fetched_data, data)

    def test_8024(self):
        "8024 - fetch blob"
        data = [(b"test_8024",)]
        self.__check_interop()
        ora_df = self.conn.fetch_df_all(
            "select to_blob(utl_raw.cast_to_raw('test_8024')) from dual"
        )
        fetched_tab = pyarrow.Table.from_arrays(
            ora_df.column_arrays(), names=ora_df.column_names()
        )
        fetched_df = fetched_tab.to_pandas()
        fetched_data = self.__get_data_from_df(fetched_df)
        self.assertEqual(fetched_data, data)

    def test_8025(self):
        "8025 - fetch raw"
        data = [(b"test_8025",)]
        self.__check_interop()
        ora_df = self.conn.fetch_df_all(
            "select utl_raw.cast_to_raw('test_8025') from dual"
        )
        fetched_tab = pyarrow.Table.from_arrays(
            ora_df.column_arrays(), names=ora_df.column_names()
        )
        fetched_df = fetched_tab.to_pandas()
        fetched_data = self.__get_data_from_df(fetched_df)
        self.assertEqual(fetched_data, data)

    @unittest.skipUnless(
        test_env.get_client_version() >= (23, 1), "unsupported client"
    )
    @unittest.skipUnless(
        test_env.get_server_version() >= (23, 1), "unsupported server"
    )
    def test_8026(self):
        "8026 - fetch boolean"
        data = [(True,), (False,), (False,), (True,), (True,)]
        self.__check_interop()
        ora_df = self.conn.fetch_df_all(
            """
            select true
            union all
            select false
            union all
            select false
            union all
            select true
            union all
            select true
            """
        )
        fetched_tab = pyarrow.Table.from_arrays(
            ora_df.column_arrays(), names=ora_df.column_names()
        )
        fetched_df = fetched_tab.to_pandas()
        fetched_data = self.__get_data_from_df(fetched_df)
        self.assertEqual(fetched_data, data)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6200 - Module for testing the methods for calling stored procedures and
functions (callproc() and callfunc()) with asyncio
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_6200(self):
        "6200 - test executing a stored procedure"
        var = self.cursor.var(oracledb.NUMBER)
        results = await self.cursor.callproc("proc_Test", ("hi", 5, var))
        self.assertEqual(results, ["hi", 10, 2.0])

    async def test_6201(self):
        "6201 - test executing a stored procedure with all args keyword args"
        inout_value = self.cursor.var(oracledb.NUMBER)
        inout_value.setvalue(0, 5)
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(
            a_InOutValue=inout_value, a_InValue="hi", a_OutValue=out_value
        )
        results = await self.cursor.callproc("proc_Test", [], kwargs)
        self.assertEqual(results, [])
        self.assertEqual(inout_value.getvalue(), 10)
        self.assertEqual(out_value.getvalue(), 2.0)

    async def test_6202(self):
        "6202 - test executing a stored procedure with last arg as keyword arg"
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_OutValue=out_value)
        results = await self.cursor.callproc("proc_Test", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertEqual(out_value.getvalue(), 2.0)

    async def test_6203(self):
        "6203 - test executing a stored procedure, repeated keyword arg"
        kwargs = dict(
            a_InValue="hi", a_OutValue=self.cursor.var(oracledb.NUMBER)
        )
        with self.assertRaisesFullCode("ORA-06550"):
            await self.cursor.callproc("proc_Test", ("hi", 5), kwargs)

    async def test_6204(self):
        "6204 - test executing a stored procedure without any arguments"
        results = await self.cursor.callproc("proc_TestNoArgs")
        self.assertEqual(results, [])

    async def test_6205(self):
        "6205 - test executing a stored function"
        results = await self.cursor.callfunc(
            "func_Test", oracledb.NUMBER, ("hi", 5)
        )
        self.assertEqual(results, 7)

    async def test_6206(self):
        "6206 - test executing a stored function without any arguments"
        results = await self.cursor.callfunc(
            "func_TestNoArgs", oracledb.NUMBER
        )
        self.assertEqual(results, 712)

    async def test_6207(self):
        "6207 - test executing a stored function with wrong parameters"
        func_name = "func_Test"
        with self.assertRaisesFullCode("DPY-2007"):
            await self.cursor.callfunc(oracledb.NUMBER, func_name, ("hi", 5))
        with self.assertRaisesFullCode("ORA-06550"):
            await self.cursor.callfunc(
                func_name, oracledb.NUMBER, ("hi", 5, 7)
            )
        with self.assertRaisesFullCode("DPY-2012"):
            await self.cursor.callfunc(func_name, oracledb.NUMBER, "hi", 7)
        with self.assertRaisesFullCode("ORA-06502"):
            await self.cursor.callfunc(func_name, oracledb.NUMBER, [5, "hi"])
        with self.assertRaisesFullCode("ORA-06550"):
            await self.cursor.callfunc(func_name, oracledb.NUMBER)
        with self.assertRaisesFullCode("DPY-2012"):
            await self.cursor.callfunc(func_name, oracledb.NUMBER, 5)

    async def test_6208(self):
        "6208 - test error for keyword args with invalid type"
        kwargs = [5]
        with self.assertRaisesFullCode("DPY-2013"):
            await self.cursor.callproc("proc_Test", [], kwargs)
        with self.assertRaisesFullCode("DPY-2013"):
            await self.cursor.callfunc(
                "func_Test", oracledb.NUMBER, [], kwargs
            )

    async def test_6209(self):
        "6209 - test calling a procedure with a string > 32767 characters"
        data = "6209" * 16000
        size_var = self.cursor.var(int)
        await self.cursor.callproc("pkg_TestLobs.GetSize", [data, size_var])
        self.assertEqual(size_var.getvalue(), len(data))

    async def test_6210(self):
        "6210 - test calling a procedure with raw data > 32767 bytes"
        data = b"6210" * 16250
        size_var = self.cursor.var(int)
        await self.cursor.callproc("pkg_TestLobs.GetSize", [data, size_var])
        self.assertEqual(size_var.getvalue(), len(data))


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# tpc_switch.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for switching two phase commit
# (TPC) transactions (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------


@cython.final
cdef class TransactionSwitchMessage(Message):
    """
    Used for two-phase commit (TPC) transaction start, attach and detach.
    """
    cdef:
        uint32_t operation, flags, timeout, application_value
        bytes context
        object xid

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization.
        """
        self.function_code = TNS_FUNC_TPC_TXN_SWITCH

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        """
        Process the parameters returned by the database.
        """
        cdef:
            const char_type* ptr
            uint16_t context_len
        buf.read_ub4(&self.application_value)
        buf.read_ub2(&context_len)
        ptr = buf.read_raw_bytes(context_len)
        self.context = ptr[:context_len]

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Writes the message to the database.
        """
        cdef:
            bytes global_transaction_id, branch_qualifier, xid_bytes
            bytes internal_name = None, external_name = None
            uint32_t format_id = 0

        # acquire data to send to the server
        if self.xid is not None:
            format_id = self.xid[0]
            global_transaction_id = self.xid[1] \
                    if isinstance(self.xid[1], bytes) \
                    else self.xid[1].encode()
            branch_qualifier = self.xid[2] \
                    if isinstance(self.xid[2], bytes) \
                    else self.xid[2].encode()
            xid_bytes = global_transaction_id + branch_qualifier
            xid_bytes += bytes(128 - len(xid_bytes))
        if self.conn_impl._internal_name is not None:
            internal_name = self.conn_impl._internal_name.encode()
        if self.conn_impl._external_name is not None:
            external_name = self.conn_impl._external_name.encode()

        # write message
        self._write_function_code(buf)
        buf.write_ub4(self.operation)
        if self.context is not None:
            buf.write_uint8(1)              # pointer (transaction context)
            buf.write_ub4(len(self.context))
        else:
            buf.write_uint8(0)              # pointer (transaction context)
            buf.write_ub4(0)                # transaction context length
        if self.xid is not None:
            buf.write_ub4(format_id)
            buf.write_ub4(len(global_transaction_id))
            buf.write_ub4(len(branch_qualifier))
            buf.write_uint8(1)              # pointer (XID)
            buf.write_ub4(len(xid_bytes))
        else:
            buf.write_ub4(0)                # format id
            buf.write_ub4(0)                # global transaction id length
            buf.write_ub4(0)                # branch qualifier length
            buf.write_uint8(0)              # pointer (XID)
            buf.write_ub4(0)                # XID length
        buf.write_ub4(self.flags)
        buf.write_ub4(self.timeout)
        buf.write_uint8(1)                  # pointer (application value)
        buf.write_uint8(1)                  # pointer (return context)
        buf.write_uint8(1)                  # pointer (return context length)
        if internal_name is not None:
            buf.write_uint8(1)              # pointer (internal name)
            buf.write_ub4(len(internal_name))
        else:
            buf.write_uint8(0)              # pointer (internal name)
            buf.write_ub4(0)                # length of internal name
        if external_name is not None:
            external_name = self.conn_impl._external_name.encode()
            buf.write_uint8(1)              # pointer (external name)
            buf.write_ub4(len(external_name))
        else:
            buf.write_uint8(0)              # pointer (external name)
            buf.write_ub4(0)                # length of external name
        if self.context is not None:
            buf.write_bytes(self.context)
        if self.xid is not None:
            buf.write_bytes(xid_bytes)
        buf.write_ub4(self.application_value)
        if internal_name is not None:
            buf.write_bytes(internal_name)
        if external_name is not None:
            buf.write_bytes(external_name)



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
8300 - Module for testing AQ with JSON queues
"""

import datetime
import decimal
import threading
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    json_queue_name = "TEST_JSON_QUEUE"
    json_data = [
        [
            2.75,
            True,
            "Ocean Beach",
            b"Some bytes",
            {"keyA": 1.0, "KeyB": "Melbourne"},
            datetime.datetime(2022, 8, 1, 0, 0),
        ],
        [
            True,
            False,
            "String",
            b"Some Bytes",
            {},
            {"name": None},
            {"name": "John"},
            {"age": 30},
            {"Permanent": True},
            {
                "employee": {
                    "name": "John",
                    "age": 30,
                    "city": "Delhi",
                    "Parmanent": True,
                }
            },
            {"employees": ["John", "Matthew", "James"]},
            {
                "employees": [
                    {"employee1": {"name": "John", "city": "Delhi"}},
                    {"employee2": {"name": "Matthew", "city": "Mumbai"}},
                    {"employee3": {"name": "James", "city": "Bangalore"}},
                ]
            },
        ],
        [
            datetime.datetime.today(),
            datetime.datetime(2004, 2, 1, 3, 4, 5),
            datetime.datetime(2020, 12, 2, 13, 29, 14),
            datetime.timedelta(8.5),
            datetime.datetime(2002, 12, 13, 9, 36, 0),
            oracledb.Timestamp(2002, 12, 13, 9, 36, 0),
            datetime.datetime(2002, 12, 13),
        ],
        dict(name="John", age=30, city="New York"),
        [
            0,
            1,
            25.25,
            6088343244,
            -9999999999999999999,
            decimal.Decimal("0.25"),
            decimal.Decimal("10.25"),
            decimal.Decimal("319438950232418390.273596"),
        ],
    ]

    def __deq_in_thread(self, results):
        with test_env.get_connection() as conn:
            queue = conn.queue(self.json_queue_name, "JSON")
            queue.deqoptions.wait = 10
            props = queue.deqone()
            if props is not None:
                results.append(props.payload)
            conn.commit()

    def __verify_attr(self, obj, attrName, value):
        setattr(obj, attrName, value)
        self.assertEqual(getattr(obj, attrName), value)

    def test_8300(self):
        "8300 - test dequeuing an empty JSON queue"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)

    def test_8301(self):
        "8301 - test enqueuing and dequeuing multiple JSON messages"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        props = self.conn.msgproperties()
        for data in self.json_data:
            props.payload = data
            queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.conn.commit()
        self.assertEqual(results, self.json_data)

    @unittest.skip("awaiting fix for bug 37746852")
    def test_8302(self):
        "8302 - test dequeuing with DEQ_REMOVE_NODATA option"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[1]
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA
        props = queue.deqone()
        self.assertIsNotNone(props)
        self.assertIsNone(props.payload)

    def test_8303(self):
        "8303 - test getting/setting dequeue options attributes"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        options = queue.deqoptions
        self.__verify_attr(options, "condition", "TEST_CONDITION")
        self.__verify_attr(options, "consumername", "TEST_CONSUMERNAME")
        self.__verify_attr(options, "correlation", "TEST_CORRELATION")
        self.__verify_attr(options, "mode", oracledb.DEQ_LOCKED)
        self.__verify_attr(
            options, "navigation", oracledb.DEQ_NEXT_TRANSACTION
        )
        self.__verify_attr(options, "transformation", "TEST_TRANSFORMATION")
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)
        self.__verify_attr(options, "wait", 1287)
        self.__verify_attr(options, "msgid", b"mID")

    def test_8304(self):
        "8304 - test waiting for dequeue"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        data = self.json_data[0]
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)
        self.conn.commit()
        thread.join()
        self.assertEqual(results, [data])

    def test_8305(self):
        "8305 - test getting/setting enqueue options attributes"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        options = queue.enqoptions
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)

    def test_8306(self):
        "8306 - test getting/setting message properties attributes"
        props = self.conn.msgproperties()
        self.__verify_attr(props, "correlation", "TEST_CORRELATION")
        self.__verify_attr(props, "delay", 60)
        self.__verify_attr(props, "exceptionq", "TEST_EXCEPTIONQ")
        self.__verify_attr(props, "expiration", 30)
        self.assertEqual(props.attempts, 0)
        self.__verify_attr(props, "priority", 1)
        self.assertEqual(props.state, oracledb.MSG_READY)
        self.assertEqual(props.deliverymode, 0)

    def test_8307(self):
        "8307 - test enqueue visibility options - ENQ_ON_COMMIT"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue(self.json_queue_name, "JSON")
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)
        self.conn.commit()
        props = queue.deqone()
        self.assertIsNotNone(props)

    def test_8308(self):
        "8308 - test enqueue visibility option - ENQ_IMMEDIATE"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue(self.json_queue_name, "JSON")
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        data = props.payload
        results = data
        other_conn.commit()
        self.assertEqual(results, self.json_data[0])

    def test_8309(self):
        "8309 - test enqueue/dequeue delivery modes identical - persistent"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue(self.json_queue_name, "JSON")
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        data = props.payload
        results = data
        other_conn.commit()
        self.assertEqual(results, self.json_data[0])

    def test_8310(self):
        "8310 - test error for message with no payload"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        props = self.conn.msgproperties()
        with self.assertRaisesFullCode("DPY-2000"):
            queue.enqone(props)

    def test_8311(self):
        "8311 - verify that the msgid property is returned correctly"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        props = self.conn.msgproperties(payload=data)
        self.assertIsNone(props.msgid)
        queue.enqone(props)
        self.cursor.execute("select msgid from JSON_QUEUE_TAB")
        (actual_msgid,) = self.cursor.fetchone()
        self.assertEqual(props.msgid, actual_msgid)
        props = queue.deqone()
        self.assertEqual(props.msgid, actual_msgid)

    def test_8312(self):
        "8312 - test message props enqtime"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        self.cursor.execute("select current_timestamp from dual")
        (start_date,) = self.cursor.fetchone()
        start_date = start_date.replace(microsecond=0)
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)
        props = queue.deqone()
        self.cursor.execute("select current_timestamp from dual")
        (end_date,) = self.cursor.fetchone()
        end_date = end_date.replace(microsecond=0)
        self.assertTrue(start_date <= props.enqtime <= end_date)

    def test_8313(self):
        "8313 - test message props declared attributes"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        values = dict(
            payload=data,
            correlation="TEST_CORRELATION",
            delay=0,
            exceptionq="PYTHONTEST.TEST_EXCEPTIONQ",
            expiration=15,
            priority=1,
        )
        props = self.conn.msgproperties(**values)
        for attr_name in values:
            self.assertEqual(getattr(props, attr_name), values[attr_name])
        queue.enqone(props)
        self.conn.commit()
        prop = queue.deqone()
        for attr_name in values:
            self.assertEqual(getattr(prop, attr_name), values[attr_name])

    def test_8314(self):
        "8314 - test getting queue attributes"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        self.assertEqual(queue.name, "TEST_JSON_QUEUE")
        self.assertEqual(queue.connection, self.conn)

    def test_8315(self):
        "8315 - test getting write-only attributes"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        for options in (queue.enqoptions, queue.deqoptions):
            with self.assertRaises(AttributeError):
                options.deliverymode

    def test_8316(self):
        "8316 - test deqoption condition with priority"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        priorities = [5, 5, 5, 5, 10, 9, 9, 10, 9]
        for priority in priorities:
            data = self.json_data[0]
            props = self.conn.msgproperties(payload=data, priority=priority)
            queue.enqone(props)

        queue.deqoptions.condition = "priority = 9"
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.conn.commit()
        self.assertEqual(len(results), 3)

    def test_8317(self):
        "8317 - test deqoption correlation"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        correlations = [
            "sample",
            "sample correlation",
            "sample",
            "sample",
            "sample correlation",
        ]
        for correlation in correlations:
            data = self.json_data[0]
            props = self.conn.msgproperties(
                payload=data, correlation=correlation
            )
            queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.correlation = "sample correlation"
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.conn.commit()
        self.assertEqual(len(results), 2)

    def test_8318(self):
        "8318 - test deqoption msgid"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        data = self.json_data[0]
        props = self.conn.msgproperties(payload=data)
        queue.enqone(props)
        queue.enqone(props)
        self.conn.commit()
        msgid = props.msgid
        queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.msgid = msgid
        prop = queue.deqone()
        self.conn.commit()
        self.assertEqual(prop.msgid, msgid)

    def test_8319(self):
        "8319 - test payload_type returns the correct value"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        self.assertEqual(queue.payload_type, "JSON")

    def test_8320(self):
        "8320 - test deprecated attributes (enqOptions, deqOptions)"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        self.assertEqual(queue.enqOptions, queue.enqoptions)
        self.assertEqual(queue.deqOptions, queue.deqoptions)

    def test_8321(self):
        "8321 - test deprecated AQ methods (enqOne, deqOne)"
        data = self.json_data[0]
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        queue.enqOne(self.conn.msgproperties(payload=data))
        props = queue.deqOne()
        self.assertEqual(props.payload, data)

    def test_8322(self):
        "8322 - test wrong payload type"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        props = self.conn.msgproperties(payload="A string")
        with self.assertRaisesFullCode("DPY-2062"):
            queue.enqone(props)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# protocol.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for establishing the protoocl to
# use during the connection (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class ProtocolMessage(Message):
    cdef:
        uint8_t server_version
        uint8_t server_flags
        bytes server_compile_caps
        bytes server_runtime_caps
        bytes server_banner

    cdef int _write_message(self, WriteBuffer buf) except -1:
        buf.write_uint8(TNS_MSG_TYPE_PROTOCOL)
        buf.write_uint8(6)                  # protocol version (8.1 and higher)
        buf.write_uint8(0)                  # "array" terminator
        buf.write_str(DRIVER_NAME)
        buf.write_uint8(0)                  # NULL terminator

    cdef int _process_message(self, ReadBuffer buf,
                              uint8_t message_type) except -1:
        if message_type == TNS_MSG_TYPE_PROTOCOL:
            self._process_protocol_info(buf)
            if not buf._caps.supports_end_of_response:
                self.end_of_response = True
        else:
            Message._process_message(self, buf, message_type)

    cdef int _process_protocol_info(self, ReadBuffer buf) except -1:
        """
        Processes the response to the protocol request.
        """
        cdef:
            uint16_t num_elem, fdo_length
            Capabilities caps = buf._caps
            const char_type *fdo
            bytearray temp_array
            ssize_t ix
        buf.read_ub1(&self.server_version)
        buf.skip_ub1()                      # skip zero byte
        self.server_banner = buf.read_null_terminated_bytes()
        buf.read_uint16le(&caps.charset_id)
        buf.read_ub1(&self.server_flags)
        buf.read_uint16le(&num_elem)
        if num_elem > 0:                    # skip elements
            buf.skip_raw_bytes(num_elem * 5)
        buf.read_uint16be(&fdo_length)
        fdo = buf.read_raw_bytes(fdo_length)
        ix = 6 + fdo[5] + fdo[6]
        caps.ncharset_id = (fdo[ix + 3] << 8) + fdo[ix + 4]
        self.server_compile_caps = buf.read_bytes()
        if self.server_compile_caps is not None:
            temp_array = bytearray(self.server_compile_caps)
            caps._adjust_for_server_compile_caps(temp_array)
            if caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_23_1:
                self.conn_impl._oson_max_fname_size = 65535
        self.server_runtime_caps = buf.read_bytes()
        if self.server_runtime_caps is not None:
            temp_array = bytearray(self.server_runtime_caps)
            caps._adjust_for_server_runtime_caps(temp_array)



# -----------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5000 - Module for testing external authentication
"""

import unittest

import oracledb
import test_env


@unittest.skipIf(
    not test_env.get_external_user(),
    "external authentication not supported with this setup",
)
class TestCase(test_env.BaseTestCase):
    require_connection = False

    def __verify_connection(
        self, connection, expected_user, expected_proxy_user=None
    ):
        with connection.cursor() as cursor:
            cursor.execute(
                """
                select
                    sys_context('userenv', 'session_user'),
                    sys_context('userenv', 'proxy_user')
                from dual
                """
            )
            actual_user, actual_proxy_user = cursor.fetchone()
            self.assertEqual(actual_user, expected_user.upper())
            self.assertEqual(
                actual_proxy_user,
                expected_proxy_user and expected_proxy_user.upper(),
            )

    def test_5000(self):
        """
        5000 - test error on creating a pool with user and password specified
        and externalauth enabled
        """
        with self.assertRaisesFullCode("DPI-1032"):
            test_env.get_pool(
                min=1,
                max=2,
                increment=1,
                getmode=oracledb.POOL_GETMODE_WAIT,
                externalauth=True,
                homogeneous=False,
            )

    def test_5001(self):
        """
        5001 - test error on creating a pool without password and with user
        specified and externalauth enabled
        """
        with self.assertRaisesFullCode("DPI-1032"):
            oracledb.create_pool(
                user=test_env.get_main_user(),
                min=1,
                max=2,
                increment=1,
                getmode=oracledb.POOL_GETMODE_WAIT,
                externalauth=True,
                homogeneous=False,
            )

    def test_5002(self):
        """
        5002 - test error on creating a pool without user and with password
        specified and externalauth enabled
        """
        with self.assertRaisesFullCode("DPI-1032"):
            oracledb.create_pool(
                password=test_env.get_main_password(),
                min=1,
                max=2,
                increment=1,
                getmode=oracledb.POOL_GETMODE_WAIT,
                externalauth=True,
                homogeneous=False,
            )

    def test_5003(self):
        """
        5003 - test creating a pool with user and password specified and
        externalauth set to False
        """
        pool = test_env.get_pool(
            min=1,
            max=2,
            increment=1,
            getmode=oracledb.POOL_GETMODE_WAIT,
            externalauth=False,
            homogeneous=False,
        )
        with pool.acquire() as conn:
            self.__verify_connection(conn, test_env.get_main_user())

    def test_5004(self):
        """
        5004 - test error when connecting with user and password specified
        and externalauth enabled
        """
        with self.assertRaisesFullCode("DPI-1032"):
            oracledb.connect(
                user=test_env.get_main_user(),
                password=test_env.get_main_password(),
                dsn=test_env.get_connect_string(),
                externalauth=True,
            )

    def test_5005(self):
        """
        5005 - test error when connecting without username and with password
        specified and externalauth enabled
        """
        with self.assertRaisesFullCode("DPI-1032"):
            oracledb.connect(
                password=test_env.get_main_password(),
                dsn=test_env.get_connect_string(),
                externalauth=True,
            )

        # by default externalauth is False
        with self.assertRaisesFullCode("ORA-01017"):
            oracledb.connect(
                password=test_env.get_main_password(),
                dsn=test_env.get_connect_string(),
            )

    def test_5006(self):
        """
        5006 - test error when connecting without password and with user
        specified and externalauth enabled
        """
        with self.assertRaisesFullCode("ORA-01017"):
            oracledb.connect(
                user="[invalid_user]",
                dsn=test_env.get_connect_string(),
                externalauth=True,
            )

        # by default externalauth is False
        with self.assertRaisesFullCode("ORA-01017"):
            oracledb.connect(
                user="[invalid_user]", dsn=test_env.get_connect_string()
            )

    def test_5007(self):
        "5007 - test external authentication with invalid proxy user"
        with self.assertRaisesFullCode("DPI-1069"):
            oracledb.connect(
                user=test_env.get_main_user(),
                dsn=test_env.get_connect_string(),
                externalauth=True,
            )

        # by default externalauth is False
        with self.assertRaisesFullCode("DPY-4001"):
            oracledb.connect(
                user=test_env.get_main_user(),
                dsn=test_env.get_connect_string(),
            )

    def test_5008(self):
        """
        5008 - test creating a connection with user and password specified and
        externalauth set to False
        """
        conn = oracledb.connect(
            user=test_env.get_main_user(),
            password=test_env.get_main_password(),
            dsn=test_env.get_connect_string(),
            externalauth=False,
        )
        self.__verify_connection(conn, test_env.get_main_user())

    def test_5009(self):
        """
        5009 - test creating standalone connection with externalauth set to
        True explicitly
        """
        conn = oracledb.connect(
            dsn=test_env.get_connect_string(), externalauth=True
        )
        self.__verify_connection(conn, test_env.get_external_user())

    def test_5010(self):
        """
        5010 - test creating standalone connection with no user and password
        specified and externalauth not set
        """
        conn = oracledb.connect(dsn=test_env.get_connect_string())
        self.__verify_connection(conn, test_env.get_external_user())

    def test_5011(self):
        "5011 - test creating a pool with external authentication"
        pool = oracledb.create_pool(
            dsn=test_env.get_connect_string(),
            min=1,
            max=2,
            increment=1,
            getmode=oracledb.POOL_GETMODE_WAIT,
            externalauth=True,
            homogeneous=False,
        )
        self.assertEqual(pool.opened, 0)
        with pool.acquire() as conn:
            self.__verify_connection(conn, test_env.get_external_user())

    def test_5012(
        self,
    ):
        """
        5012 - test creating a pool without user and password specified and
        externalauth not set
        """
        pool = oracledb.create_pool(
            dsn=test_env.get_connect_string(),
            min=1,
            max=2,
            increment=1,
            getmode=oracledb.POOL_GETMODE_WAIT,
            homogeneous=False,
        )
        with self.assertRaisesFullCode("ORA-24415"):
            pool.acquire()

    def test_5013(self):
        "5013 - test pool min is always 0 under external authentication"
        pool = oracledb.create_pool(
            dsn=test_env.get_connect_string(),
            min=5,
            max=10,
            increment=3,
            getmode=oracledb.POOL_GETMODE_WAIT,
            externalauth=True,
            homogeneous=False,
        )
        self.assertEqual(pool.opened, 0)

    def test_5014(self):
        "5014 - test pool increment is always 1 under external authentication"
        pool = oracledb.create_pool(
            dsn=test_env.get_connect_string(),
            min=5,
            max=10,
            increment=3,
            getmode=oracledb.POOL_GETMODE_WAIT,
            externalauth=True,
            homogeneous=False,
        )
        conn1 = pool.acquire()
        self.assertEqual(pool.opened, 1)
        conn2 = pool.acquire()
        self.assertEqual(pool.opened, 2)
        conn1.close()
        conn2.close()

    def test_5015(self):
        "5015 - test external authentication with proxy"
        proxy_user = test_env.get_external_user()  # proxy user
        schema_user = test_env.get_main_user()  # schema user
        conn1 = oracledb.connect(
            user=f"[{schema_user}]",
            dsn=test_env.get_connect_string(),
            externalauth=True,
        )
        self.__verify_connection(conn1, schema_user, proxy_user)
        conn2 = oracledb.connect(
            user=f"[{schema_user}]", dsn=test_env.get_connect_string()
        )
        self.__verify_connection(conn2, schema_user, proxy_user)

    def test_5016(self):
        "5016 - test creating pool using external authentication with proxy"
        proxy_user = test_env.get_external_user()
        schema_user = test_env.get_main_user()
        pool = oracledb.create_pool(
            externalauth=True,
            homogeneous=False,
            dsn=test_env.get_connect_string(),
            min=1,
            max=2,
            increment=1,
            getmode=oracledb.POOL_GETMODE_WAIT,
        )
        self.assertEqual(pool.opened, 0)
        conn = pool.acquire(user=f"[{schema_user}]")
        self.__verify_connection(conn, schema_user, proxy_user)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# utils.py
#
# Contains utility classes and methods.
# -----------------------------------------------------------------------------

from typing import Callable, Union

from . import base_impl
from . import driver_mode
from . import errors


def enable_thin_mode():
    """
    Makes python-oracledb be in Thin mode. After this method is called, Thick
    mode cannot be enabled. If python-oracledb is already in Thick mode, then
    calling ``enable_thin_mode()`` will fail. If connections have already been
    opened, or a connection pool created, in Thin mode, then calling
    ``enable_thin_mode()`` is a no-op.

    Since python-oracledb defaults to Thin mode, almost all applications do not
    need to call this method. However, because it bypasses python-oracledb's
    internal mode-determination heuristic, it may be useful for applications
    that are using standalone connections in multiple threads to concurrently
    create connections when the application starts.
    """
    with driver_mode.get_manager(requested_thin_mode=True):
        pass


def params_initer(f):
    """
    Decorator function which is used on the ConnectParams and PoolParams
    classes. It creates the implementation object using the implementation
    class stored on the parameter class. It first, however, calls the original
    method to ensure that the keyword parameters supplied are valid (the
    original method itself does nothing).
    """

    def wrapped_f(self, *args, **kwargs):
        f(self, *args, **kwargs)
        self._impl = self._impl_class()
        if kwargs:
            self._impl.set(kwargs)

    return wrapped_f


def params_setter(f):
    """
    Decorator function which is used on the ConnectParams and PoolParams
    classes. It calls the set() method on the parameter implementation object
    with the supplied keyword arguments. It first, however, calls the original
    method to ensure that the keyword parameters supplied are valid (the
    original method itself does nothing).
    """

    def wrapped_f(self, *args, **kwargs):
        f(self, *args, **kwargs)
        self._impl.set(kwargs)

    return wrapped_f


def register_params_hook(hook_function: Callable) -> None:
    """
    Registers a user function to be called internally prior to connection or
    pool creation. The hook function accepts a copy of the parameters that will
    be used to create the pool or standalone connection and may modify them.
    For example, the cloud native authentication plugins modify the
    "access_token" parameter with a function that will acquire the token using
    information found in the "extra_auth_parms" parameter.
    """
    if hook_function is None or not callable(hook_function):
        raise TypeError("hook_function must be a callable and cannot be None")
    base_impl.REGISTERED_PARAMS_HOOKS.append(hook_function)


def register_password_type(
    password_type: str, hook_function: Callable
) -> None:
    """
    Registers a user function to be called when a password is provided as a
    dictionary containing a key "type" with the specified value. The hook
    function is expected to use the dictionary and return the password value.
    If the supplied function is None, the registration is removed.
    """
    if not isinstance(password_type, str):
        raise TypeError("password_type must be a string")
    if hook_function is not None and not callable(hook_function):
        raise TypeError("hook_function must be a callable")
    password_type = password_type.lower()
    if hook_function is None:
        base_impl.REGISTERED_PASSWORD_TYPES.pop(password_type)
    else:
        base_impl.REGISTERED_PASSWORD_TYPES[password_type] = hook_function


def register_protocol(protocol: str, hook_function: Callable) -> None:
    """
    Registers a user function to be called prior to connection or pool creation
    when an Easy Connect connection string prefixed with the specified protocol
    is being parsed internally by python-oracledb in Thin mode. The registered
    function will also be invoked by ConnectParams.parse_connect_string() in
    Thin and Thick modes. Your hook function is expected to find or construct a
    valid connection string. If the supplied function is None, the registration
    is removed.
    """
    if not isinstance(protocol, str):
        raise TypeError("protocol must be a string")
    if hook_function is not None and not callable(hook_function):
        raise TypeError("hook_function must be a callable")
    protocol = protocol.lower()
    if hook_function is None:
        base_impl.REGISTERED_PROTOCOLS.pop(protocol)
    else:
        base_impl.REGISTERED_PROTOCOLS[protocol] = hook_function


def unregister_params_hook(hook_function: Callable) -> None:
    """
    Unregisters a user function that was earlier registered with a call to
    register_params_hook().
    """
    base_impl.REGISTERED_PARAMS_HOOKS.remove(hook_function)


def verify_stored_proc_args(
    parameters: Union[list, tuple], keyword_parameters: dict
) -> None:
    """
    Verifies that the arguments to a call to a stored procedure or function
    are acceptable.
    """
    if parameters is not None and not isinstance(parameters, (list, tuple)):
        errors._raise_err(errors.ERR_ARGS_MUST_BE_LIST_OR_TUPLE)
    if keyword_parameters is not None and not isinstance(
        keyword_parameters, dict
    ):
        errors._raise_err(errors.ERR_KEYWORD_ARGS_MUST_BE_DICT)



# -----------------------------------------------------------------------------
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
4400 - Module for testing TPC (two-phase commit) transactions.
"""

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def test_4400(self):
        "4400 - test begin, prepare, roll back global transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3900, b"txn3900", b"branchId")
        self.conn.tpc_begin(xid)
        self.assertEqual(self.conn.tpc_prepare(), False)
        self.conn.tpc_begin(xid)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.assertEqual(self.conn.tpc_prepare(), True)
        self.conn.tpc_rollback()
        self.cursor.execute("select count(*) from TestTempTable")
        (count,) = self.cursor.fetchone()
        self.assertEqual(count, 0)

    def test_4401(self):
        "4401 - test begin, prepare, commit global transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3901, "txn3901", "branchId")
        self.conn.tpc_begin(xid)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.assertEqual(self.conn.tpc_prepare(), True)
        self.conn.tpc_commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(), [(1, "tesName")])

    def test_4402(self):
        "4402 - test multiple global transactions on the same connection"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3902, "txn3902", "branch1")
        xid2 = self.conn.xid(3902, b"txn3902", b"branch2")
        self.conn.tpc_begin(xid1)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.conn.tpc_end()
        self.conn.tpc_begin(xid2)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'tesName')
            """
        )
        self.conn.tpc_end()
        needs_commit1 = self.conn.tpc_prepare(xid1)
        needs_commit2 = self.conn.tpc_prepare(xid2)
        if needs_commit1:
            self.conn.tpc_commit(xid1)
        if needs_commit2:
            self.conn.tpc_commit(xid2)
        self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        expected_rows = [(1, "tesName"), (2, "tesName")]
        self.assertEqual(self.cursor.fetchall(), expected_rows)

    def test_4403(self):
        "4403 - test rollback with parameter xid"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3901, b"txn3901", b"branch1")
        xid2 = self.conn.xid(3902, "txn3902", "branch2")
        for count, xid in enumerate([xid1, xid2]):
            self.conn.tpc_begin(xid)
            self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:id, 'tesName')
                """,
                id=count,
            )
            self.conn.tpc_end()
        self.conn.tpc_rollback(xid1)

        with self.assertRaisesFullCode("ORA-24756"):
            self.conn.tpc_prepare(xid1)
        needs_commit = self.conn.tpc_prepare(xid2)
        if needs_commit:
            self.conn.tpc_commit(xid2)
        self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        self.assertEqual(self.cursor.fetchall(), [(1, "tesName")])

    def test_4404(self):
        "4404 - test resuming a transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3939, "txn3939", "branch39")
        xid2 = self.conn.xid(3940, "txn3940", "branch40")
        values = [[xid1, (1, "User Info")], [xid2, (2, "Other User Info")]]
        for xid, data in values:
            self.conn.tpc_begin(xid)
            self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                """,
                data,
            )
            self.conn.tpc_end()
        for xid, data in values:
            self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_RESUME)
            self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
            (res,) = self.cursor.fetchall()
            self.assertEqual(res, data)
            self.conn.tpc_rollback(xid)

    def test_4405(self):
        "4405 - test promoting a local transaction to a tpc transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3941, "txn3941", "branch41")
        values = (1, "String 1")
        self.cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)",
            values,
        )
        with self.assertRaisesFullCode("ORA-24776"):
            self.conn.tpc_begin(xid)
        self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_PROMOTE)
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        (res,) = self.cursor.fetchall()
        self.assertEqual(res, values)
        self.conn.tpc_rollback(xid)

    def test_4406(self):
        "4406 - test ending a transaction with parameter xid"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(4406, "txn4406a", "branch3")
        xid2 = self.conn.xid(4406, b"txn4406b", b"branch4")
        self.conn.tpc_begin(xid1)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test4406a')
            """
        )
        self.conn.tpc_begin(xid2)
        with self.assertRaisesFullCode("ORA-24758"):
            self.conn.tpc_end(xid1)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'test4406b')
            """
        )
        self.conn.tpc_end(xid2)
        with self.assertRaisesFullCode("ORA-25352"):
            self.conn.tpc_end(xid1)
        self.conn.tpc_rollback(xid1)
        self.conn.tpc_rollback(xid2)

    def test_4407(self):
        "4407 - test tpc_recover()"
        self.cursor.execute("truncate table TestTempTable")
        n_xids = 10
        for i in range(n_xids):
            xid = self.conn.xid(4407 + i, f"txn4407{i}", f"branch{i}")
            self.conn.tpc_begin(xid)
            self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, 'test4407')
                """,
                [i + 1],
            )
            self.conn.tpc_prepare(xid)
        recovers = self.conn.tpc_recover()
        self.assertEqual(len(recovers), n_xids)

        self.cursor.execute("select * from DBA_PENDING_TRANSACTIONS")
        self.assertEqual(self.cursor.fetchall(), recovers)

        for xid in recovers:
            if xid.format_id % 2 == 0:
                self.conn.tpc_commit(xid)
        recovers = self.conn.tpc_recover()
        self.assertEqual(len(recovers), n_xids // 2)

        for xid in recovers:
            self.conn.tpc_rollback(xid)
        recovers = self.conn.tpc_recover()
        self.assertEqual(len(recovers), 0)

    def test_4408(self):
        "4408 - test tpc_recover() with read-only transaction"
        self.cursor.execute("truncate table TestTempTable")
        for i in range(4):
            xid = self.conn.xid(4408 + i, f"txn4408{i}", f"branch{i}")
            self.conn.tpc_begin(xid)
            self.cursor.execute("select * from TestTempTable")
            self.conn.tpc_prepare(xid)
        recovers = self.conn.tpc_recover()
        self.assertEqual(len(recovers), 0)

    def test_4409(self):
        "4409 - test tpc_commit() with one_phase parameter"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4409, "txn4409", "branch1")
        self.conn.tpc_begin(xid)
        values = (1, "test4409")
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            values,
        )
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.conn.tpc_commit(xid, one_phase=True)
        self.assertEqual(self.cursor.fetchall(), [values])

    def test_4410(self):
        "4410 - test negative cases for tpc_commit()"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4410, "txn4410", "branch1")
        self.conn.tpc_begin(xid)
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test4410')
            """
        )
        self.assertRaises(TypeError, self.conn.tpc_commit, "invalid xid")
        self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-02053"):
            self.conn.tpc_commit(xid, one_phase=True)
        with self.assertRaisesFullCode("ORA-24756"):
            self.conn.tpc_commit(xid)

    def test_4411(self):
        "4411 - test starting an already created transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4411, "txn4411", "branch1")
        self.conn.tpc_begin(xid)
        with self.assertRaisesFullCode("ORA-24757"):
            self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_NEW)
        with self.assertRaisesFullCode("ORA-24797"):
            self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_PROMOTE)
        self.conn.tpc_end()
        for flag in [oracledb.TPC_BEGIN_NEW, oracledb.TPC_BEGIN_PROMOTE]:
            with self.assertRaisesFullCode("ORA-24757"):
                self.conn.tpc_begin(xid, flag)
        self.conn.tpc_rollback(xid)

    def test_4412(self):
        "4412 - test resuming a prepared transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4412, "txn4412", "branch1")
        self.conn.tpc_begin(xid)
        self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-24756"):
            self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_RESUME)

    def test_4413(self):
        "4413 - test tpc_begin and tpc_end with invalid parameters"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4413, "txn4413", "branch1")
        test_values = [
            (self.conn.tpc_begin, "DPY-2050"),
            (self.conn.tpc_end, "DPY-2051"),
        ]
        for tpc_function, error_code in test_values:
            self.assertRaises(TypeError, tpc_function, "invalid xid")
            with self.assertRaisesFullCode(error_code):
                tpc_function(xid, "invalid flag")
            with self.assertRaisesFullCode(error_code):
                tpc_function(xid, 70)

    def test_4414(self):
        "4414 - test commiting transaction without tpc_commit"
        xid = self.conn.xid(4414, "txn4409", "branch1")
        self.conn.tpc_begin(xid)
        with self.assertRaisesFullCode("ORA-02089"):
            self.cursor.execute("truncate table TestTempTable")

    def test_4415(self):
        "4415 - test tpc_commit when a commit is not needed"
        xid = self.conn.xid(4416, "txn4416", "branch1")
        self.conn.tpc_begin(xid)
        self.cursor.execute("select * from TestTempTable")
        self.conn.tpc_end(xid)
        self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-24756"):
            self.conn.tpc_commit(xid)

    def test_4416(self):
        "4416 - test transaction_in_progress"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(4415, "txn4415", "branch1")
        self.assertFalse(self.conn.transaction_in_progress)

        self.conn.tpc_begin(xid)
        self.assertTrue(self.conn.transaction_in_progress)
        self.cursor.execute("insert into TestTempTable (IntCol) values (2)")

        self.conn.tpc_end(xid)
        self.assertFalse(self.conn.transaction_in_progress)

        self.conn.tpc_prepare(xid)
        self.assertFalse(self.conn.transaction_in_progress)

        self.conn.tpc_commit(xid)
        self.assertFalse(self.conn.transaction_in_progress)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# execute.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for executing a SQL statement
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class ExecuteMessage(MessageWithData):
    cdef:
        uint32_t fetch_orientation
        uint32_t fetch_pos
        bint scroll_operation

    cdef int _write_execute_message(self, WriteBuffer buf) except -1:
        """
        Write the message for a full execute.
        """
        cdef:
            uint32_t options, exec_flags = 0, num_params = 0, num_iters = 1
            Statement stmt = self.cursor_impl._statement
            BaseThinCursorImpl cursor_impl = self.cursor_impl
            list params = stmt._bind_info_list

        # determine the options to use for the execute
        options = 0
        if not stmt._requires_define and not self.parse_only \
                and params is not None:
            num_params = <uint32_t> len(params)
        if stmt._requires_define:
            options |= TNS_EXEC_OPTION_DEFINE
        elif not self.parse_only and stmt._sql is not None:
            exec_flags |= TNS_EXEC_FLAGS_IMPLICIT_RESULTSET
            if not self.scroll_operation:
                options |= TNS_EXEC_OPTION_EXECUTE
        if cursor_impl.scrollable:
            exec_flags |= TNS_EXEC_FLAGS_SCROLLABLE
        if stmt._cursor_id == 0 or stmt._is_ddl:
            options |= TNS_EXEC_OPTION_PARSE
        if stmt._is_query:
            if self.parse_only:
                options |= TNS_EXEC_OPTION_DESCRIBE
            else:
                if stmt._cursor_id == 0 or stmt._requires_define:
                    num_iters = self.cursor_impl.prefetchrows
                else:
                    num_iters = self.cursor_impl.arraysize
                self.cursor_impl._set_fetch_array_size(num_iters)
                if num_iters > 0 and not stmt._no_prefetch:
                    options |= TNS_EXEC_OPTION_FETCH
        if not stmt._is_plsql and not self.parse_only:
            options |= TNS_EXEC_OPTION_NOT_PLSQL
        elif stmt._is_plsql and num_params > 0:
            options |= TNS_EXEC_OPTION_PLSQL_BIND
        if num_params > 0:
            options |= TNS_EXEC_OPTION_BIND
        if self.batcherrors:
            options |= TNS_EXEC_OPTION_BATCH_ERRORS
        if self.arraydmlrowcounts:
            exec_flags |= TNS_EXEC_FLAGS_DML_ROWCOUNTS
        if self.conn_impl.autocommit and not self.parse_only:
            options |= TNS_EXEC_OPTION_COMMIT

        # write body of message
        self._write_function_code(buf)
        buf.write_ub4(options)              # execute options
        buf.write_ub4(stmt._cursor_id)      # cursor id
        if stmt._cursor_id == 0 or stmt._is_ddl:
            buf.write_uint8(1)              # pointer (cursor id)
            buf.write_ub4(stmt._sql_length)
        else:
            buf.write_uint8(0)              # pointer (cursor id)
            buf.write_ub4(0)
        buf.write_uint8(1)                  # pointer (vector)
        buf.write_ub4(13)                   # al8i4 array length
        buf.write_uint8(0)                  # pointer (al8o4)
        buf.write_uint8(0)                  # pointer (al8o4l)
        buf.write_ub4(0)                    # prefetch buffer size
        buf.write_ub4(num_iters)            # prefetch number of rows
        buf.write_ub4(TNS_MAX_LONG_LENGTH)  # maximum long size
        if num_params == 0:
            buf.write_uint8(0)              # pointer (binds)
            buf.write_ub4(0)                # number of binds
        else:
            buf.write_uint8(1)              # pointer (binds)
            buf.write_ub4(num_params)       # number of binds
        buf.write_uint8(0)                  # pointer (al8app)
        buf.write_uint8(0)                  # pointer (al8txn)
        buf.write_uint8(0)                  # pointer (al8txl)
        buf.write_uint8(0)                  # pointer (al8kv)
        buf.write_uint8(0)                  # pointer (al8kvl)
        if stmt._requires_define:
            buf.write_uint8(1)              # pointer (al8doac)
            buf.write_ub4(len(self.cursor_impl.fetch_vars))
                                            # number of defines
        else:
            buf.write_uint8(0)
            buf.write_ub4(0)
        buf.write_ub4(0)                    # registration id
        buf.write_uint8(0)                  # pointer (al8objlist)
        buf.write_uint8(1)                  # pointer (al8objlen)
        buf.write_uint8(0)                  # pointer (al8blv)
        buf.write_ub4(0)                    # al8blvl
        buf.write_uint8(0)                  # pointer (al8dnam)
        buf.write_ub4(0)                    # al8dnaml
        buf.write_ub4(0)                    # al8regid_msb
        if self.arraydmlrowcounts:
            buf.write_uint8(1)              # pointer (al8pidmlrc)
            buf.write_ub4(self.num_execs)   # al8pidmlrcbl
            buf.write_uint8(1)              # pointer (al8pidmlrcl)
        else:
            buf.write_uint8(0)              # pointer (al8pidmlrc)
            buf.write_ub4(0)                # al8pidmlrcbl
            buf.write_uint8(0)              # pointer (al8pidmlrcl)
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_12_2:
            buf.write_uint8(0)                  # pointer (al8sqlsig)
            buf.write_ub4(0)                    # SQL signature length
            buf.write_uint8(0)                  # pointer (SQL ID)
            buf.write_ub4(0)                    # allocated size of SQL ID
            buf.write_uint8(0)                  # pointer (length of SQL ID)
            if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_12_2_EXT1:
                buf.write_uint8(0)              # pointer (chunk ids)
                buf.write_ub4(0)                # number of chunk ids
        if stmt._cursor_id == 0 or stmt._is_ddl:
            if stmt._sql_bytes is None:
                errors._raise_err(errors.ERR_INVALID_REF_CURSOR)
            buf.write_bytes_with_length(stmt._sql_bytes)
            buf.write_ub4(1)                # al8i4[0] parse
        else:
            buf.write_ub4(0)                # al8i4[0] parse
        if stmt._is_query:
            if stmt._cursor_id == 0:
                buf.write_ub4(0)            # al8i4[1] execution count
            else:
                buf.write_ub4(num_iters)
        else:
            buf.write_ub4(self.num_execs)   # al8i4[1] execution count
        buf.write_ub4(0)                    # al8i4[2]
        buf.write_ub4(0)                    # al8i4[3]
        buf.write_ub4(0)                    # al8i4[4]
        buf.write_ub4(0)                    # al8i4[5] SCN (part 1)
        buf.write_ub4(0)                    # al8i4[6] SCN (part 2)
        buf.write_ub4(stmt._is_query)       # al8i4[7] is query
        buf.write_ub4(0)                    # al8i4[8]
        buf.write_ub4(exec_flags)           # al8i4[9] execute flags
        buf.write_ub4(self.fetch_orientation) # al8i4[10] fetch orientation
        buf.write_ub4(self.fetch_pos)       # al8i4[11] fetch pos
        buf.write_ub4(0)                    # al8i4[12]
        if stmt._requires_define:
            self._write_column_metadata(buf, self.cursor_impl.fetch_var_impls)
        elif num_params > 0:
            self._write_bind_params(buf, params)

    cdef int _write_reexecute_message(self, WriteBuffer buf) except -1:
        """
        Write the message for a re-execute.
        """
        cdef:
            uint32_t i, options_1 = 0, options_2 = 0, num_iters
            Statement stmt = self.cursor_impl._statement
            list params = stmt._bind_info_list
            BindInfo info

        if params:
            if not stmt._is_query and not stmt._is_returning:
                self.out_var_impls = [info._bind_var_impl \
                                      for info in params \
                                      if info.bind_dir != TNS_BIND_DIR_INPUT]
            params = [info for info in params \
                      if info.bind_dir != TNS_BIND_DIR_OUTPUT \
                      and not info._is_return_bind]
        if self.function_code == TNS_FUNC_REEXECUTE_AND_FETCH:
            options_1 |= TNS_EXEC_OPTION_EXECUTE
            num_iters = self.cursor_impl.prefetchrows
            self.cursor_impl._set_fetch_array_size(num_iters)
        else:
            if self.conn_impl.autocommit:
                options_2 |= TNS_EXEC_OPTION_COMMIT_REEXECUTE
            num_iters = self.num_execs

        self._write_function_code(buf)
        buf.write_ub4(stmt._cursor_id)
        buf.write_ub4(num_iters)
        buf.write_ub4(options_1)
        buf.write_ub4(options_2)
        if params:
            for i in range(self.num_execs):
                buf.write_uint8(TNS_MSG_TYPE_ROW_DATA)
                self._write_bind_params_row(buf, params, i)

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Write the execute message to the buffer. Two types of execute messages
        are possible: one for a full execute and the second, simpler message,
        for when an existing cursor is being re-executed. A full execute is
        required under the following circumstances:
            - the statement has never been executed
            - the statement refers to a REF cursor (no sql is defined)
            - prefetch is not possible (LOB columns fetched)
            - bind metadata has changed
            - parse is being performed
            - define is being performed
            - DDL is being executed
            - batch errors mode is enabled
        """
        cdef:
            Statement stmt = self.cursor_impl._statement
        if stmt._cursor_id == 0 or not stmt._executed \
                or stmt._sql is None \
                or stmt._no_prefetch \
                or stmt._binds_changed \
                or self.parse_only \
                or stmt._requires_define \
                or stmt._is_ddl \
                or self.batcherrors \
                or self.cursor_impl.scrollable:
            self.function_code = TNS_FUNC_EXECUTE
            self._write_execute_message(buf)
        elif stmt._is_query and self.cursor_impl.prefetchrows > 0:
            self.function_code = TNS_FUNC_REEXECUTE_AND_FETCH
            self._write_reexecute_message(buf)
        else:
            self.function_code = TNS_FUNC_REEXECUTE
            self._write_reexecute_message(buf)
        stmt._binds_changed = False

    cdef int process(self, ReadBuffer buf) except -1:
        """
        Runs after the database response has been processed. If the statement
        executed requires define and is not a REF cursor (which would already
        have performed the define during its execute), then mark the message as
        needing to be resent. If this is after the second time the message has
        been sent, mark the statement as no longer needing a define (since this
        only needs to happen once).
        """
        cdef Statement stmt = self.cursor_impl._statement
        MessageWithData.process(self, buf)
        if self.error_occurred and self.error_info.pos == 0 and stmt._is_plsql:
            self.error_info.pos = self.error_info.rowcount + self.offset
        if not self.parse_only:
            stmt._executed = True
        if stmt._requires_define and stmt._sql is not None:
            if self.resend:
                stmt._requires_define = False
            else:
                self.resend = True



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5200 - Module for testing the SQL parser.
"""

import unittest

import test_env


class TestCase(test_env.BaseTestCase):
    def test_5200(self):
        "5200 - single line comment"
        self.cursor.prepare(
            "--begin :value2 := :a + :b + :c +:a +3; end;\n"
            "begin :value2 := :a + :c +3; end; -- not a :bind_variable"
        )
        self.assertEqual(self.cursor.bindnames(), ["VALUE2", "A", "C"])

    def test_5201(self):
        "5201 - multiple line comment"
        self.cursor.prepare(
            "/*--select * from :a where :a = 1\n"
            "select * from table_names where :a = 1*/\n"
            "select :table_name, :value from dual"
        )
        self.assertEqual(self.cursor.bindnames(), ["TABLE_NAME", "VALUE"])

    def test_5202(self):
        "5202 - constant strings"
        statement = """
                    begin
                        :value := to_date('20021231 12:31:00', :format);
                    end;"""
        self.cursor.prepare(statement)
        self.assertEqual(self.cursor.bindnames(), ["VALUE", "FORMAT"])

    def test_5203(self):
        "5203 - multiple division operators"
        self.cursor.prepare(
            """
            select :a / :b, :c / :d
            from dual
            """
        )
        self.assertEqual(self.cursor.bindnames(), ["A", "B", "C", "D"])

    def test_5204(self):
        "5204 - starting with parentheses"
        sql = "(select :a from dual) union (select :b from dual)"
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["A", "B"])

    def test_5205(self):
        "5205 - invalid quoted bind"
        sql = 'select ":test", :a from dual'
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["A"])

    def test_5206(self):
        "5206 - non-ascii character in the bind name"
        sql = "select :méil$ from dual"
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["MÉIL$"])

    def test_5207(self):
        "5207 - various quoted bind names"
        tests = [
            ('select :"percent%" from dual', ["percent%"]),
            ('select : "q?marks" from dual', ["q?marks"]),
            ('select :"percent%(ens)yah" from dual', ["percent%(ens)yah"]),
            ('select :  "per % cent" from dual', ["per % cent"]),
            ('select :"per cent" from dual', ["per cent"]),
            ('select :"par(ens)" from dual', ["par(ens)"]),
            ('select :"more/slashes" from dual', ["more/slashes"]),
            ('select :"%percent" from dual', ["%percent"]),
            ('select :"/slashes/" from dual', ["/slashes/"]),
            ('select :"1col:on" from dual', ["1col:on"]),
            ('select :"col:ons" from dual', ["col:ons"]),
            ('select :"more :: %colons%" from dual', ["more :: %colons%"]),
            ('select :"more/slashes" from dual', ["more/slashes"]),
            ('select :"spaces % spaces" from dual', ["spaces % spaces"]),
            ('select "col:nns", :"col:ons", :id from dual', ["col:ons", "ID"]),
        ]
        for sql, expected in tests:
            with self.subTest(sql=sql, expected=expected):
                self.cursor.prepare(sql)
                self.assertEqual(self.cursor.bindnames(), expected)

    def test_5208(self):
        "5208 - sql containing quoted identifiers and strings"
        sql = 'select "/*_value1" + : "VaLue_2" + :"*/3VALUE" from dual'
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["VaLue_2", "*/3VALUE"])

    def test_5209(self):
        "5209 - statement containing simple strings"
        sql = """select '"string_1"', :bind_1, ':string_2' from dual"""
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["BIND_1"])

    def test_5210(self):
        "5210 - bind variables between comment blocks"
        self.cursor.prepare(
            """
            select
                /* comment 1 with /* */
                :a,
                /* comment 2 with another /* */
                :b
                /* comment 3 * * * / */,
                :c
            from dual
            """
        )
        self.assertEqual(self.cursor.bindnames(), ["A", "B", "C"])

    def test_5211(self):
        "5211 - bind variables between q-strings"
        self.cursor.prepare(
            """
            select
                :a,
                q'{This contains ' and " and : just fine}',
                :b,
                q'[This contains ' and " and : just fine]',
                :c,
                q'<This contains ' and " and : just fine>',
                :d,
                q'(This contains ' and " and : just fine)',
                :e,
                q'$This contains ' and " and : just fine$',
                :f
            from dual
            """
        )
        self.assertEqual(
            self.cursor.bindnames(), ["A", "B", "C", "D", "E", "F"]
        )

    @unittest.skipUnless(
        test_env.get_client_version() >= (19, 1), "unsupported client"
    )
    def test_5212(self):
        "5212 - bind variables between JSON constants"
        self.cursor.prepare(
            """
            select
                json_object('foo':dummy),
                :bv1,
                json_object('foo'::bv2),
                :bv3,
                json { 'key1': 57, 'key2' : 58 },
                :bv4
            from dual
            """
        )
        self.assertEqual(self.cursor.bindnames(), ["BV1", "BV2", "BV3", "BV4"])

    def test_5213(self):
        "5213 - multiple line comment with multiple asterisks"
        self.cursor.prepare(
            "/****--select * from :a where :a = 1\n"
            "select * from table_names where :a = 1****/\n"
            "select :table_name, :value from dual"
        )
        self.assertEqual(self.cursor.bindnames(), ["TABLE_NAME", "VALUE"])

    def test_5214(self):
        "5214 - qstring without a closing quote"
        with self.assertRaisesFullCode("DPY-2041"):
            self.cursor.prepare("select q'[something from dual")

    def test_5215_different_space_combinations(self):
        "5215 - different space combinations with :="
        self.cursor.prepare(
            """
            begin :value2 :=
                               :a  + :b  +   :c +:a +3; end;
                               begin :value2
                               :=
                               :a + :c +3; end;
            """
        )
        self.assertEqual(self.cursor.bindnames(), ["VALUE2", "A", "B", "C"])

    def test_5216_binds_between_comment_blocks_with_quotes(self):
        "5216 - bind variables between multiple comment blocks with quotes"
        self.cursor.prepare(
            """
            select
                /* ' comment 1 */
                :a,
                /* "comment " 2 ' */:b
                /* comment 3 '*/,
                :c
                /* comment 4 ""*/
            from dual
            """
        )
        self.assertEqual(self.cursor.bindnames(), ["A", "B", "C"])

    def test_5217_binds_missing_quote_exception(self):
        "5217 - query with a missing end quote"
        with self.assertRaisesFullCode("DPY-2041"):
            self.cursor.prepare("select 'abc, :a from dual")

    def test_5218_binds_missing_quote_exception(self):
        "5218 - q-string with wrong closing symbols"
        with self.assertRaisesFullCode("DPY-2041"):
            self.cursor.prepare("select q'[abc'], 5 from dual")


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E2300 - Module for testing Transaction Guard (TG). No special setup is required
but the test suite makes use of debugging packages that are not intended for
normal use. It also creates and drops a service.
"""

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    service_name = "oracledb-test-tg"
    requires_connection = False

    @classmethod
    def setUpClass(cls):
        cls.admin_conn = test_env.get_admin_connection()
        user = test_env.get_main_user()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(
                f"""
                declare
                    params          dbms_service.svc_parameter_array;
                begin
                    params('COMMIT_OUTCOME') := 'true';
                    params('RETENTION_TIMEOUT') := 604800;
                    dbms_service.create_service('{cls.service_name}',
                                                '{cls.service_name}', params);
                    dbms_service.start_service('{cls.service_name}');
                end;
                """
            )
            cursor.execute(f"grant execute on dbms_tg_dbg to {user}")
            cursor.execute(f"grant execute on dbms_app_cont to {user}")

    @classmethod
    def tearDownClass(cls):
        user = test_env.get_main_user()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(f"revoke execute on dbms_tg_dbg from {user}")
            cursor.execute(f"revoke execute on dbms_app_cont from {user}")
            cursor.callproc("dbms_service.stop_service", [cls.service_name])
            cursor.callproc("dbms_service.delete_service", [cls.service_name])

    def test_ext_2300(self):
        "E2300 - test standalone connection"
        params = test_env.get_connect_params().copy()
        params.parse_connect_string(test_env.get_connect_string())
        params.set(service_name=self.service_name)
        for arg_name in ("pre_commit", "post_commit"):
            with self.subTest(arg_name=arg_name):
                conn = oracledb.connect(params=params)
                cursor = conn.cursor()
                cursor.execute("truncate table TestTempTable")
                cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)
                    """,
                    [2300, "String for test 2300"],
                )
                full_arg_name = f"dbms_tg_dbg.tg_failpoint_{arg_name}"
                cursor.execute(
                    f"""
                    begin
                        dbms_tg_dbg.set_failpoint({full_arg_name});
                    end;
                    """
                )
                ltxid = conn.ltxid
                with self.assertRaisesFullCode("DPY-4011"):
                    conn.commit()
                conn = oracledb.connect(params=params)
                cursor = conn.cursor()
                committed_var = cursor.var(bool)
                completed_var = cursor.var(bool)
                cursor.callproc(
                    "dbms_app_cont.get_ltxid_outcome",
                    [ltxid, committed_var, completed_var],
                )
                expected_value = arg_name == "post_commit"
                self.assertEqual(committed_var.getvalue(), expected_value)
                self.assertEqual(completed_var.getvalue(), expected_value)

    def test_ext_2301(self):
        "E2301 - test pooled connection"
        params = test_env.get_pool_params().copy()
        params.parse_connect_string(test_env.get_connect_string())
        params.set(service_name=self.service_name, max=10)
        pool = oracledb.create_pool(params=params)
        for arg_name in ("pre_commit", "post_commit"):
            with self.subTest(arg_name=arg_name):
                conn = pool.acquire()
                cursor = conn.cursor()
                cursor.execute("truncate table TestTempTable")
                cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)
                    """,
                    [2300, "String for test 2300"],
                )
                full_arg_name = f"dbms_tg_dbg.tg_failpoint_{arg_name}"
                cursor.execute(
                    f"""
                    begin
                        dbms_tg_dbg.set_failpoint({full_arg_name});
                    end;
                    """
                )
                ltxid = conn.ltxid
                with self.assertRaisesFullCode("DPY-4011"):
                    conn.commit()
                conn = pool.acquire()
                cursor = conn.cursor()
                committed_var = cursor.var(bool)
                completed_var = cursor.var(bool)
                cursor.callproc(
                    "dbms_app_cont.get_ltxid_outcome",
                    [ltxid, committed_var, completed_var],
                )
                expected_value = arg_name == "post_commit"
                self.assertEqual(committed_var.getvalue(), expected_value)
                self.assertEqual(completed_var.getvalue(), expected_value)


if __name__ == "__main__":
    test_env.run_test_cases()
